@@ -1,7 +1,7 @@
 /**
  * \file
  * \author Guillaume Papin <guillaume.papin@epitech.eu>
- * 
+ *
  * \brief Command parser definitions.
  *
  * This file is distributed under the GNU General Public License. See
@@ -16,6 +16,8 @@
 #include <cstdlib>
 #include <functional>
 #include <iostream>
+#include <map>
+
 
 namespace {
 
@@ -81,6 +83,43 @@ private:
   bool *dest_;
 };
 
+const std::map<std::string, PrefixMatchStyle> PREFIX_MATCH_STYLE_MAP = {
+  { "exact", PrefixMatchStyle::Exact },
+  { "case-insensitive", PrefixMatchStyle::CaseInsensitive },
+  { "smart-case", PrefixMatchStyle::SmartCase},
+};
+
+/// Convert style to a PrefixMatchStyle
+struct PrefixMatchStyleConverter {
+  PrefixMatchStyleConverter(PrefixMatchStyle *dest) : dest_(dest) {
+  }
+
+  bool operator()(const std::string &str) {
+    auto res = PREFIX_MATCH_STYLE_MAP.find(str);
+
+    if (res == PREFIX_MATCH_STYLE_MAP.cend()) {
+      return false;
+    }
+    *dest_ = res->second;
+    return true;
+  }
+
+private:
+  PrefixMatchStyle *dest_;
+};
+
+std::ostream &operator<<(std::ostream &os, PrefixMatchStyle style) {
+  for (auto it : PREFIX_MATCH_STYLE_MAP) {
+    if (it.second == style) {
+      os << it.first;
+      return os;
+    }
+  }
+  os << "UnknownStyle";
+  return os;
+}
+
+
 } // unnamed namespace
 
 std::ostream &operator<<(std::ostream &os, const Command::Action &action) {
@@ -99,9 +138,12 @@ std::ostream &operator<<(std::ostream &os, const Command::Action &action) {
 std::ostream &operator<<(std::ostream &os, const Command &command) {
   os << "Command{action=" << command.action << ", "
      << "file='" << command.file << "', "
+     << "unsavedFile='" << command.unsavedFile << "', "
      << "dir='" << command.dir << "', "
      << "line=" << command.line << ", "
      << "column=" << command.column << ", "
+     << "prefix='" << command.prefix << "', "
+     << "caseStyle='" << command.style << "', "
      << "flags=[";
   bool first = true;
   for (const std::string &flag : command.flags) {
@@ -111,7 +153,6 @@ std::ostream &operator<<(std::ostream &os, const Command &command) {
     first = false;
   }
   os << "], "
-     << "unsavedFiles.count=" << command.unsavedFiles.size() << ", "
      << "opt=" << (command.opt ? "on" : "off");
 
   return os << "}";
@@ -143,7 +184,6 @@ Command *CommandParser::parse(const std::vector<std::string> &argv) {
 
   command_.action = actionFromString(actionStr);
 
-  bool handleUnsaved = false;
   bool readCompileOptions = false;
   std::vector<std::function<bool(const std::string &)>> positionalArgs;
 
@@ -154,7 +194,6 @@ Command *CommandParser::parse(const std::vector<std::string> &argv) {
 
   case Command::Parse:
     positionalArgs.push_back(StringConverter(&command_.file));
-    handleUnsaved = true;
     readCompileOptions = true;
     break;
 
@@ -162,7 +201,6 @@ Command *CommandParser::parse(const std::vector<std::string> &argv) {
     positionalArgs.push_back(StringConverter(&command_.file));
     positionalArgs.push_back(UnsignedIntConverter(&command_.line));
     positionalArgs.push_back(UnsignedIntConverter(&command_.column));
-    handleUnsaved = true;
     readCompileOptions = true;
     break;
 
@@ -171,6 +209,20 @@ Command *CommandParser::parse(const std::vector<std::string> &argv) {
     positionalArgs.push_back(UnsignedIntConverter(&command_.column));
     break;
 
+  case Command::SetUnsaved:
+    positionalArgs.push_back(StringConverter(&command_.file));
+    positionalArgs.push_back(StringConverter(&command_.unsavedFile));
+    break;
+
+  case Command::ResetUnsaved:
+    positionalArgs.push_back(StringConverter(&command_.file));
+    break;
+
+  case Command::Candidates:
+    positionalArgs.push_back(StringConverter(&command_.prefix));
+    positionalArgs.push_back(PrefixMatchStyleConverter(&command_.style));
+    break;
+  case Command::CompletionDiagnostics:
   case Command::Diagnostics:
   case Command::Help:
   case Command::Exit:
@@ -187,31 +239,13 @@ Command *CommandParser::parse(const std::vector<std::string> &argv) {
     return 0;
   }
 
-  auto argIt = argv.begin() + 1;
-  int argCount = std::distance(argIt, argv.end());
+  auto argsBegin = argv.begin() + 1;
+  const auto argsEnd = std::find(argsBegin, argv.end(), "--");
+  const int argCount = std::distance(argsBegin, argsEnd);
 
-  // parse optional arguments come first
-  while (argIt != argv.end()) {
-    // '-' is allowed as a "default" file, this isn't an option but a positional
-    // argument
-    if ((*argIt)[0] != '-' || *argIt == "-")
-      break;
-
-    const std::string &opt = *argIt;
-
-    ++argIt;
-    argCount--;
-
-    if (handleUnsaved) {
-      // TODO: handle multiple unsaved files
-      if (opt == "--num-unsaved=1") {
-        command_.unsavedFiles.resize(1);
-      }
-    } else {
-      std::clog << "error: invalid option for '" << actionStr << "': '" << opt
-                << "' unknown\n";
-      return 0;
-    }
+  // compile options are provided after '--'
+  if (readCompileOptions && argsEnd != argv.end()) {
+    command_.flags.assign(std::next(argsEnd), argv.end());
   }
 
   if (argCount != static_cast<int>(positionalArgs.size())) {
@@ -222,67 +256,22 @@ Command *CommandParser::parse(const std::vector<std::string> &argv) {
   }
 
   for (auto fn : positionalArgs) {
-    if (!fn(*argIt)) {
+    if (!fn(*argsBegin)) {
       std::clog << "error: parsing command '" << actionStr
-                << "': invalid argument '" << *argIt << "'\n";
+                << "': invalid argument '" << *argsBegin << "'\n";
       return 0;
     }
-    ++argIt;
+    ++argsBegin;
   }
 
   // '-' is used as a special file to inform that the buffer hasn't been saved
   // on disk and only the buffer content is available. libclang needs a file, so
   // this is treated as a special value for irony-server to create a temporary
-  // file for this. note taht libclang will gladly accept '-' as a filename but
+  // file for this. note that libclang will gladly accept '-' as a filename but
   // we don't want to let this happen since irony already reads stdin.
   if (command_.file == "-") {
     command_.file = tempFile_.getPath();
   }
 
-  // When a file is provided, the next line contains the compilation options to
-  // pass to libclang.
-  if (readCompileOptions) {
-    std::string compileOptions;
-    std::getline(std::cin, compileOptions);
-
-    command_.flags = unescapeCommandLine(compileOptions);
-  }
-
-  // read unsaved files
-  // filename
-  // filesize
-  // <file content...>
-  for (auto &p : command_.unsavedFiles) {
-    std::getline(std::cin, p.first);
-
-    unsigned length;
-    std::string filesizeStr;
-    std::getline(std::cin, filesizeStr);
-
-    UnsignedIntConverter uintConverter(&length);
-
-    if (!uintConverter(filesizeStr)) {
-      std::clog << "error: invalid file size '" << filesizeStr << "'\n";
-      return 0;
-    }
-
-    p.second.resize(length);
-    std::cin.read(p.second.data(), p.second.size());
-
-    CXUnsavedFile cxUnsavedFile;
-
-    cxUnsavedFile.Filename = p.first.c_str();
-    cxUnsavedFile.Contents = p.second.data();
-    cxUnsavedFile.Length = p.second.size();
-    command_.cxUnsavedFiles.push_back(cxUnsavedFile);
-
-    char nl;
-    std::cin.read(&nl, 1);
-    if (nl != '\n') {
-      std::clog << "error: missing newline for unsaved file content\n";
-      return 0;
-    }
-  }
-
   return &command_;
 }
